@@ -1,8 +1,9 @@
 #  Copyright (c) 2022 Szymon Mikler
 
-from . import graph_algorithms, model_tools, symbolic_data
-from .config import optimize_module_calls
+
+from . import config, graph_algorithms, model_tools, symbolic_data, useful_layers
 from .functions_utility import add_to_graph
+from .symbolic_api_2 import SymbolicAPI2ContextManager, optimize_module_calls
 from .symbolic_data import CustomInput, Input
 from .symbolic_model import SymbolicModel
 
@@ -11,8 +12,11 @@ __all__ = [
     "CustomInput",
     "SymbolicModel",
     "add_to_graph",
+    "config",
     "graph_algorithms",
     "model_tools",
     "optimize_module_calls",
     "symbolic_data",
+    "useful_layers",
+    "SymbolicAPI2ContextManager",
 ]
